"""
    SecTheory.LocalOptimizer.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.NumericalUtils import safe_ratios
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy._MOLASS.SerialSettings import get_setting

PENALTY_SCALE = 1e5
NUM_SECCOL_PARAMS = 6
MEMP_LIMIT = 3.1
N_LIMIT = 3000

def split_params(p, nc):
        return (
            p[0:NUM_SECCOL_PARAMS],
            p[NUM_SECCOL_PARAMS:NUM_SECCOL_PARAMS+nc], 
            p[NUM_SECCOL_PARAMS+nc:NUM_SECCOL_PARAMS+nc*2]
            )

class RgInfo:
    def __init__(self, rg_curve, nc):
        trs, _, rgs = rg_curve.get_valid_curves()
        self.mask = rg_curve.get_mask()
        self.rg_weights = rg_curve.get_weights()

        mean = np.mean(rgs)
        sigma = np.std(rgs)

        # self.rg_lower = min(mean - 2*sigma, np.min(rgs))/2
        self.rg_lower = 3
        self.rg_upper = max(mean + 2*sigma, np.max(rgs))*2
        print("----------- self.rg_lower=", self.rg_lower)
        print("----------- self.rg_upper=", self.rg_upper)
        self.trs = trs
        self.rgs = rgs
        self.zeros_rg = np.zeros(nc-1)
        self.ones_rgs = np.ones(len(rgs))

    def get_order_penalty(self, rg):
        return np.sum(np.min([self.zeros_rg, rg[:-1] - rg[1:]], axis=0)**2)

    def get_bound_penalty(self, rg):
        return min(0, np.min(rg) - self.rg_lower)**2 + min(0, self.rg_upper - np.max(rg))**2

class SecInfo:
    def __init__(self, t0_upper_bound=None):
        self.d_init = get_setting("poresize")
        rp_lower, rp_upper = get_setting("poresize_bounds")
        self.d_lower = rp_lower
        self.d_upper = rp_upper
        self.t0_upper_bound = t0_upper_bound
        self.t0_init = 0 if t0_upper_bound is None else t0_upper_bound*0.7
        self.N = 500

    def get_sec_params(self, K):
        N = self.N
        return self.t0_init, self.d_init, N, 1, K/N, 1

class LocalOptimizer:
    def __init__(self, ecurve, rg_info, sec_info, nc, elutionmodel_func, logger=None):
        self.ecurve = ecurve
        self.rg_info = rg_info
        self.sec_info = sec_info
        self.nc = nc
        self.elutionmodel_func = elutionmodel_func
        if logger is None:
            import logging
            logger = logging.getLogger(__name__)
        self.logger = logger
        self.init_tr = None
        self.abort = False

    def objective_all(self, p, get_curves=False, debug=False, debug_title="all"):
        ecurve = self.ecurve
        x = ecurve.x
        y = ecurve.y
        max_y = ecurve.max_y

        rgs = self.rg_info.rgs

        d_init = self.sec_info.d_init
        d_lower = self.sec_info.d_lower
        d_upper = self.sec_info.d_upper
        t0_init = self.sec_info.t0_init

        (t0, rp, N, me, T, mp), rg, w = split_params(p, self.nc)
        if debug:
            print("(t0, rp, N, me, T, mp)=", (t0, rp, N, me, T, mp))
            print("rg=", rg)
            print("w=", w)

        rho = rg/rp
        rho[rho > 1] = 1
        ty = np.zeros(len(x))

        model_trs = t0 + N*T*(1 - rho)**(me + mp)

        if self.init_tr is None:
            tr_decency = 0
        else:
            tr_decency = np.sum((model_trs - self.init_tr)**2)

        penalty = min(0, N)**2 + min(0, T)**2 

        cy_list = []
        for w_, r_ in zip(w, rho):
            penalty += min(0, w_)**2
            np_ = N*(1 - r_)**me
            tp_ = T*(1 - r_)**mp
            cy = w_*self.elutionmodel_func(x - t0, np_, tp_)
            ty += cy
            cy_list.append(cy)

        if get_curves:
            return cy_list, ty

        penalty += self.rg_info.get_order_penalty(rg)
        penalty += self.rg_info.get_bound_penalty(rg)
        penalty += min(0, rp - d_lower)**2 + min(0, d_upper - rp)**2
        penalty += min(0, MEMP_LIMIT - (me+mp))**2 + min(0, me)**2 + min(0, mp)**2
        t0_upper_bound = self.sec_info.t0_upper_bound
        if t0_upper_bound is not None:
            penalty += max(0, t0 - t0_upper_bound)**2

        mask = self.rg_info.mask

        tym = ty[mask]
        t_rg = np.zeros(len(tym))

        for r, cy in zip(rg, cy_list):
            t_rg += r * safe_ratios(self.rg_info.ones_rgs, cy[mask], tym)

        xr_fitting = np.log10(np.sum((ty - y)**2))
        if not np.isfinite(xr_fitting):
            xr_fitting = PENALTY_SCALE
            penalty += 1
        rg_fitting = 0.1*np.log10(np.average((self.rg_info.rg_weights*(t_rg - rgs)/t_rg)**2))
        if not np.isfinite(rg_fitting):
            rg_fitting = PENALTY_SCALE
            penalty += 1
        penalty *= PENALTY_SCALE
        fv = xr_fitting + rg_fitting + penalty + tr_decency
        if np.isnan(fv):
            fv = penalty

        if debug:
            print("xr_fitting=", xr_fitting)
            print("rg_fitting=", rg_fitting)
            print("penalty=", penalty)
            print("fv=", fv)
            rg_info = self.rg_info
            trs = rg_info.trs
            rg_upper = rg_info.rg_upper

            in_folder = get_in_folder()

            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("%s: Decomposition %s" % (debug_title, in_folder), fontsize=16)
                axt = ax.twinx()
                axt.grid(False)
                ax.plot(x, y, label="data")
                axt.plot(trs, rgs, color="gray", alpha=0.5, label="observed rg")
                axt.plot(trs, t_rg, ":", color="black", label="reconstructed rg")
                axt.set_ylim(0, rg_upper)

                print("t0, rp, N, me, T, mp=", t0, rp, N, me, T, mp)
                print("rg=", rg)
                print("w=", w)

                for k, (cy, tr) in enumerate(zip(cy_list, model_trs)):
                    ax.plot(x, cy, ":", label="component-%d" % (k+1))
                    if np.isfinite(tr):
                        j = int(tr)
                        if j >= 0 and j < len(cy):
                            ax.plot(tr, cy[j], "o", color="yellow")
                ax.plot(x, ty, ":", color="red", label="component-total")

                ax.legend(fontsize=14)
                axt.legend(loc="upper left", fontsize=14)
                fig.tight_layout()
                ret = plt.show()
            if not ret:
                self.abort = True

        return fv

    def optimize(self, init_params, params_given=False, init_tr=None, global_opt=False, debug=False):
        self.init_tr = init_tr
        if global_opt:
            if debug:
                self.objective_all(init_params, debug=True)

            ret = basinhopping(self.objective_all, init_params)
        else:
            init_fv = self.objective_all(init_params, debug=debug)
            print("init_fv=", init_fv)

            if params_given:
                ret = minimize(self.objective_all, init_params)
                self.logger.info("estimate_monopore_params: minimize success=%s, status=%d, message=%s", str(ret.success), ret.status, ret.message)
            else:
                k = 0
                # (may be default) method="BFGS" seems best
                # , method="Nelder-Mead"
                ret = minimize(self.objective_all, init_params)
                print([k], "ret.fun=", ret.fun, "ret.success=", ret.success, "ret.message=", ret.message)

                if ret.success:
                    self.logger.info("estimate_monopore_params: minimize[%d] ok with fv=%g", k, ret.fun)
                    # break
                else:
                    self.logger.info("estimate_monopore_params: minimize[%d] failed with fv=%g, status=%d, message=%s", k, ret.fun, ret.status, ret.message)

        if debug:
            self.objective_all(ret.x, debug=True)

        return ret



"""
    Optimizer.FullOptResult.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import os
import re
import logging
import numpy as np
from molass_legacy.Trimming import restore_trimming_info_impl
from molass_legacy.Baseline.BaselineUtils import create_xr_baseline_object
from .StateSequence import read_callback_txt_impl, read_seed_txt
from .FuncImporter import import_objective_function
from molass_legacy.SecSaxs.DataTreatment import DataTreatment
from .OptJobInfo import OptJobInfo
from .OptDataSets import OptDataSets
from .OptLogFile import OptLogFile
from .TheUtils import FILES

OPTIONAL_FILES = [
    "composite_info.txt",
    ]


class FullOptResult:
    def __init__(self, orig_sd, pre_recog, folder, set_folder_setting=False):
        self.logger = logging.getLogger(__name__)
        self.logger.info("constructing a result object from %s", folder)
        self.orig_sd = orig_sd          # not yet trimmed
        self.pre_recog = pre_recog
        # self.folder = folder.replace("/", "\\") # internal path delimiter should be \?
        self.folder = folder
        self.optimizer_folder = self.get_optimizer_folder_impl()
        if set_folder_setting:
            # temporary fix for Batch.BatchUtils.load_v2_result
            from molass_legacy._MOLASS.SerialSettings import set_setting
            set_setting('optimizer_folder', self.optimizer_folder)
        trimming_txt = os.path.join(folder, FILES[6])
        restore_trimming_info_impl(trimming_txt, self.logger)
        self.load_info()
        self.best_index = None

    def load_composite(self):
        composite_info_txt = os.path.join(self.folder, OPTIONAL_FILES[0])
        if os.path.exists(composite_info_txt):
            from .CompositeInfo import CompositeInfo
            composite = CompositeInfo()
            composite.load(composite_info_txt)
        else:
            composite = None
        return composite

    def get_log_file(self):
        optimizer_log = os.path.join(self.folder, FILES[3])
        return OptLogFile(optimizer_log)

    def get_class_code(self, log_file):
        self.class_code =  log_file.class_code
        return self.class_code

    def load_callback_info(self):
        callback_txt = os.path.join(self.folder, FILES[0])
        fv_list, x_list = read_callback_txt_impl(callback_txt)
        fv_array = np.array([rec[1] for rec in fv_list])
        return fv_list, x_list, fv_array
    
    def load_init_params(self):
        init_params_txt = os.path.join(self.folder, FILES[2])
        return np.loadtxt(init_params_txt)

    def load_info(self, debug=False):
        from .OptimizerSettings import OptimizerSettings

        optimizer_folder = self.get_optimizer_folder()

        self.settings = settings = OptimizerSettings()
        settings.load(optimizer_folder=optimizer_folder)    # must be loaded before fullopt construction

        self.treat = treat = DataTreatment()
        treat.load(optimizer_folder=optimizer_folder)
        # self.sd_copy = treat.get_treated_sd(self.orig_sd, self.pre_recog)
        self.sd_copy = treat.get_trimmed_sd(self.orig_sd, self.pre_recog)
        corrected_sd = treat.get_corrected_sd(self.orig_sd, self.pre_recog, self.sd_copy)
        uv_base_curve = treat.get_base_curve()

        self.dsets = OptDataSets(self.sd_copy, corrected_sd, possibly_relocated=True, current_folder=self.folder)
        composite = self.load_composite()
        log_file = self.get_log_file()
        class_code = self.get_class_code(log_file)
        self.fullopt_class = import_objective_function(class_code, self.logger)
        self.drift_type = "linear"

        info = OptJobInfo()
        info.load(self.folder)
        self.n_components = info.nc

        xr_base_curve = create_xr_baseline_object()
        self.fullopt = self.fullopt_class(self.dsets, self.n_components,
                            uv_base_curve=uv_base_curve,
                            xr_base_curve=xr_base_curve,
                            qvector=self.sd_copy.qvector,   # sd_copy: trimmed_sd
                            wvector=self.sd_copy.lvector,
                            composite=composite
                            )

        if self.fullopt.is_stochastic():
            from molass_legacy.Estimators.SdmEstimatorProxy import SdmEstimatorProxy
            estimator = SdmEstimatorProxy(self.folder)
            self.fullopt.params_type.set_estimator(estimator)      # reconsider the neccesity of this line

        fv_list, x_list, fv_array= self.load_callback_info()
        init_params = self.load_init_params()

        if log_file.version_date < "2022-09-20":

            if not debug:
                import molass_legacy.KekLib.CustomMessageBox as MessageBox
                MessageBox.showerror("Not Supported Error",
                    "View of results made with versions earlier than 2022-09-20\n"
                    "is not supported!",
                    parent=None,
                    )
                assert False

            adjuster = self.fullopt.params_type.get_adjuster()
            adjuster.fit(self.fullopt, fv_array, x_list)
            init_params = adjuster.convert(init_params)
            x_list = adjuster.convert(x_list)

        self.init_params = init_params
        self.fv_list = fv_list
        self.fv_array = fv_array
        self.x_list = x_list
        self.n_iterations = len(fv_list) - 2

        seed_txt = os.path.join(self.folder, FILES[5])
        self.seed = read_seed_txt(seed_txt)
        self.composite = composite

        bounds_txt = os.path.join(self.folder, FILES[7])
        if os.path.exists(bounds_txt):
            real_bounds = np.loadtxt(bounds_txt)
        else:
            real_bounds = None

        self.fullopt.prepare_for_optimization(self.init_params, real_bounds=real_bounds)

        self.logger.info("load_info: n_components=%d, len(init_params)=%d, composite=%s", self.n_components, len(self.init_params), str(composite))

    def get_optimizer(self):
        return self.fullopt

    def get_optimizer_folder_impl(self):
        return os.path.dirname(os.path.dirname(self.folder))

    def get_optimizer_folder(self):
        return self.optimizer_folder

    def get_optimizer_jobs_folder(self):
        return os.path.join(self.optimizer_folder, "jobs")

    def get_jobnames(self):
        nodes = os.listdir(self.get_optimizer_jobs_folder())
        return nodes

    def get_num_jobs(self):
        nodes = self.get_jobnames()
        return len(nodes)

    def get_rg_folder(self):
        return os.path.join(self.optimizer_folder, "rg-curve")

    def get_init_info(self):
        from molass_legacy.Optimizer.InitialInfo import InitialInfo
        return InitialInfo(result=self)

    def get_demo_index(self):
        return self.n_iterations

    def get_job_name(self):
        import re
        job_name_re = re.compile(r"(\d+)$")
        m = job_name_re.search(self.folder)
        if m:
            name = m.group(1)
        else:
            name = "???"
        return name

    def get_job_list(self):
        jobs_folder = self.get_optimizer_jobs_folder()
        job_list = []
        for name in self.get_jobnames():
            try:
                folder = os.path.join(jobs_folder, name)
                info = OptJobInfo()
                info.load(folder)
                job_list.append(info)
            except:
                # may be the folder is not complete: missing "optimizer.log", etc.
                from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                log_exception(self.logger, "get_job_list: ")
        return job_list

    def get_best_params(self):
        if self.best_index is None:
            self.best_index = np.argmin(self.fv_array)
        return self.x_list[self.best_index]

    def get_result_iterator(self, all=False):
        from molass_legacy.Optimizer.ParamsIterator import create_iterator
        if all:
            for i, params in enumerate(self.x_list):
                yield i, params
        else:
            for i in create_iterator(self.fv_array):
                yield i, self.x_list[i]

    def debug_plot(self, params):
        optimizer = self.get_optimizer()
        # def objective_func(self, p, plot=False, debug=False, fig_info=None, axis_info=None, return_full=False):
        optimizer.objective_func(params, plot=True)

def get_result_from_workfolder(workfolder, replace_specs=None):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.LiteBatch import LiteBatch
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition

    in_data_path = os.path.join(workfolder, FILES[1])
    with open(in_data_path, "r") as f:
        in_folder = f.read().strip().split('=')[1]

    if replace_specs is not None:
        for k, v in replace_specs:
            in_folder = in_folder.replace(k, v)

    analysis_folder = '\\'.join(workfolder.split('\\')[:-3])

    print("analysis_folder=", analysis_folder)
    print("workfolder=", workfolder)
    print("in_folder=", in_folder)
    set_setting("in_folder", in_folder)
    set_setting("analysis_folder", analysis_folder)

    batch = LiteBatch()
    sd = batch.load_data(in_folder)
    pre_recog = PreliminaryRecognition(sd)
    result = FullOptResult(sd, pre_recog, workfolder)
    return result


"""
    BraceWidgets.py

    Copyright (c) 2022-2023, Masatsuyo Takahashi
"""
import os
import sys
import tkinter as Tk
from PIL import Image, ImageTk

# ClosingBrace or RightBrace
class ClosingBrace(Tk.Label):
    def __init__( self, parent, **kwargs ):
        dir_, _ = os.path.split(__file__)
        image = Image.open(os.path.join( dir_, "closing_brace.png"))
        w, h = image.size
        w_ = kwargs.pop("width", w)
        h_ = kwargs.pop("height", h)
        self.image = ImageTk.PhotoImage(image.resize((w_, h_)))
        Tk.Label.__init__( self, parent, image=self.image, compound=Tk.LEFT, **kwargs )

if __name__ == '__main__':
    root = Tk.Tk()
    frame = Tk.Frame( root )
    frame.pack( padx=30, pady=10 )
    brace = ClosingBrace(frame, height=50)
    # brace = ClosingBrace(frame)
    brace.pack()

    root.mainloop()



# coding: utf-8
"""
    RgProcess.RgManager.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""

class RgManager:
    def __init__(self, in_folder, sd):
        import multiprocessing as mp
        si = self.get_sharedinfo(in_folder, sd)
        self.process = mp.Process(target=self.run, args=(si,))
        self.process.start()

    def get_sharedinfo(self, in_folder, sd):
        from Processes.SharedInfo import SharedInfo
        si = SharedInfo()
        si.in_folder.set(in_folder)
        si.initilize_buffer(array=sd.intensity_array, ecurve=sd.get_xray_curve())
        return si

    def run(self, si):
        from molass_legacy.KekLib.TkUtils import get_tk_root
        from .RgAnalysis import RgAnalysisDialog

        si.get_buffer_ready()
        root = get_tk_root()

        def show_dialog():
            dialog  = RgAnalysisDialog(root, si)
            dialog.show()
            root.quit()

        root.after(0, show_dialog)
        root.mainloop()
        root.destroy()



# coding: utf-8
"""
    MakeItMoreClever.py

    Copyright (c) 2019-2021, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import is_empty_val
from molass_legacy._MOLASS.SerialSettings import get_setting

def increase_menu_availability(parent):
    an_folder = parent.an_folder.get()
    analysis_name = parent.analysis_name.get()
    analysis_folder = get_setting('analysis_folder')

    if not is_empty_val(an_folder) and not is_empty_val(analysis_name) and analysis_folder is None:
        parent.new_analysis()
        parent.menu4.update_states()



"""
    NaiiveLrfSolver.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from SvdDenoise import get_denoised_data

class NaiiveLrfSolver:
    def __init__(self, qv, D, E, i=None, denoise=True):
        if i is None:
            from bisect import bisect_right
            i = bisect_right(qv, 0.02)

        self.qv = qv
        self.i = i
        if denoise:
            D_ = get_denoised_data(D, rank=2)
        else:
            D_ = D
        self.D_ = D_
        self.E = E
        cy = np.average(D_[i-5:i+6,:], axis=0)
        self.j = np.argmax(cy)
        self.Cinit = np.array([cy, cy**2])

    def solve(self, debug=False):
        qv = self.qv
        i = self.i
        j = self.j
        C = self.Cinit
        D_ = self.D_
        Cinv = np.linalg.pinv(C)
        P = D_ @ Cinv
        return P, C



"""

    ProgressInfo.py

    Copyright (c) 2017-2023, Masatsuyo Takahashi, KEK-PF

"""
import queue
import time

INFO_TYPE_STEP  = 0
INFO_TYPE_MAX   = 1

STATE_INIT      = 0
STATE_DOING     = 1
STATE_DONE      = 2
SAFE_FINISH     = 3
STATE_ERROR     = -1
STATE_FATAL     = -2

exe_queue  = queue.Queue()
cmd_queue  = queue.Queue()

def cleanup_queues():
    for q in [ exe_queue, cmd_queue ]:
        while not q.empty():
            q.get()

"""
    stream      phase       max
    Guinier     0   autorg  num_files
                1   excel   openpyxl(), com()

    Zero-Ex     0   asc(0)  
                1   desc(0) 
                2   asc(1)  
                3   desc(1) 
                .
                .   merge   com()
"""

class StreamState:
    def __init__( self, phase, step, state ):
        self.phase  = phase
        self.step   = step
        self.state  = state

class ExecInfo:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

        """
        self.info_type  :   0 step change
                            1 max change

        self.stream     :   0 
                            1 
        self.phase      :   0, 1, 2
        self.step       :   0, 1, 2, ..., num_steps
        self.state      :   
        self.message    :   
        """

    def __str__( self ):
        return str( self.state )

class ProgressInfo:
    def __init__( self, init_max_dict ):
        self.stream_max_dict        = {}
        self.stream_cur_dict        = {}
        self.stream_end_phase_dict  = {}
        self.g_cumulative_max_dict  = {}
        self.stream_phases          = {}
        self.stream_time_dict       = {}
        self.g_max_dict             = init_max_dict
        for k, v in sorted( self.g_max_dict.items() ):
            stream  = k[0]
            phase   = k[1]
            smax = self.stream_max_dict.get( stream )
            if smax is None:
                smax = v
            else:
                smax += v
            self.g_cumulative_max_dict[k]       = smax
            self.stream_max_dict[stream]        = smax
            self.stream_end_phase_dict[stream]  = phase

        self.stream_state_dict = {}
        for stream in self.stream_max_dict.keys():
            self.stream_state_dict[stream]  = STATE_INIT
            self.stream_cur_dict[stream]    = 0
            self.stream_phases[stream]      = {}

        self.current_stream = 0

    def get_start_time( self, stream ):
        rec = self.stream_time_dict.get(stream)
        return None if rec is None else rec[0]

    def get_info( self, block=False ):
        """
            note that get_info is called only in the main thread
            so that there is no need to care for thread safety.
        """

        try:
            info = exe_queue.get( block )
        except queue.Empty:
            info = None

        if info is None: return info

        if info.info_type == INFO_TYPE_MAX:
            new_progress_info = ProgressInfo( info.max_dict )
            # assumed this info is given prior to info.stream step info
            # and info.max_dict includes info.stream's max info only
            self.stream_max_dict[info.stream]       = new_progress_info.stream_max_dict[info.stream]
            self.stream_end_phase_dict[info.stream] = new_progress_info.stream_end_phase_dict[info.stream]
            assert( self.stream_state_dict[info.stream] == STATE_INIT )
            assert( self.stream_cur_dict[info.stream] == 0 )
            for k, v in info.max_dict.items():
                self.g_max_dict[k]              = v
                self.g_cumulative_max_dict[k]   = new_progress_info.g_cumulative_max_dict[k]
            print( 'max info changed by', info.max_dict )
            return

        if info.phase >= 0:
            self.stream_phases[info.stream][info.phase] = 1
            time_rec = self.stream_time_dict.get(info.stream)
            if time_rec is None:
                self.stream_time_dict[info.stream] = [ time.time(), None ]

        if info.state is None:
            end_phase = self.stream_end_phase_dict[info.stream]
            if info.phase >= end_phase:
                num_steps = self.g_max_dict.get( (info.stream, info.phase) )
                if num_steps is None:
                    # TODO
                    assert( False )
                if info.step >= num_steps:
                    state = STATE_DONE
                    time_rec = self.stream_time_dict.get(info.stream)
                    time_rec[1] = time.time()
                else:
                    state = STATE_DOING
            else:
                state = STATE_DOING
        else:
            # info.state is treated as stronger info than info.step
            state = info.state

        if info.phase == 0:
            cumulative_num_steps    = 0
        else:
            try:
                last_phase = sorted( self.stream_phases[info.stream].keys() )[-2]
                cumulative_num_steps    = self.g_cumulative_max_dict[ (info.stream, last_phase ) ]
            except IndexError as exc:
                # print( self.stream_phases )
                # print( self.g_cumulative_max_dict )
                # raise exc
                cumulative_num_steps    = 0

        if info.step >= 0:
            self.stream_cur_dict[info.stream]    = cumulative_num_steps + max( 0, info.step )
        self.stream_state_dict[info.stream]  = state
        self.current_stream = info.stream

        return info

    def get_stream_states( self ):
        return [ self.stream_state_dict[stream]
                    for stream in range( len(self.stream_state_dict) ) ]

    def get_stream_progresses( self ):
        return [ ( self.stream_max_dict[stream], self.stream_cur_dict[stream], self.stream_state_dict[stream] )
                    for stream in range( len(self.stream_state_dict) ) ]

    def is_end_states( self, states ):
        is_end = True
        for stream, state in self.stream_state_dict.items():
            max_ = self.stream_max_dict[stream]
            if max_ == 0:
                # if there is nothing todo, it is ok to be regarded as done
                continue

            if state == STATE_INIT or state == STATE_DOING:
                is_end = False
            elif state == STATE_ERROR:
                cur_ = self.stream_cur_dict[stream]
                # print( 'cur_, max_', cur_, max_ )
                if cur_ < max_:
                    # In cases of STATE_ERROR
                    # it is not taken as finished if there remain things to do.
                    # On the other hand, STATE_FATAL means imediately "is_end"
                    is_end = False
        return is_end

    def set_error( self ):
        put_error( self.current_stream )

def put_info( k, step, state=None ):
    stream, phase = k
    info = ExecInfo( info_type=INFO_TYPE_STEP, stream=stream, phase=phase, step=step, state=state )
    exe_queue.put( info )

def put_error( k, step=-1, error_state=STATE_ERROR ):
    if type( k ) == tuple:
        stream, phase = k
    else:
        stream = k
        phase  = -1
    info = ExecInfo( info_type=INFO_TYPE_STEP, stream=stream, phase=phase, step=step, state=error_state )
    exe_queue.put( info )

def put_max_info( stream, max_dict ):
    # check whether max_dict is restricted to stream
    for k, v in max_dict.items():
        assert( k[0] == stream )

    info = ExecInfo( info_type=INFO_TYPE_MAX, stream=stream, max_dict=max_dict )
    exe_queue.put( info )

def send_stop():
    cmd_queue.put( 'STOP' )

def on_stop_raise( cleanup=None, log_closure=None ):
    try:
        cmd = cmd_queue.get( False )
        if log_closure is not None:
            log_closure(cmd)
    except:
        cmd = None

    if cmd is not None:
        if cleanup is not None:
            cleanup()
        raise Exception( 'STOP' )



"""
    Optimizer.OptimalEgh.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
from time import time
import numpy as np
from scipy import optimize
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from SvdDenoise import get_denoised_data
from molass_legacy.Models.ElutionCurveModels import (EGHA, EGH, egh, egh_x_from_height_ratio,
                                EMGA, EMG, emg, emg_x_from_height_ratio)
from MatrixData import simple_plot_3d

# PARAMS = ['h', 'mu', 'sigma', 'tau']    # 'a' not used
PARAMS = np.arange(4)
NUM_PARAMS = len(PARAMS)
MAX_ITERATION_FMIN_CG = 100
MAX_ITERATION_FMIN = 1000
VERY_SMALL_VALUE = 1e-6
USE_FACTOED_WIEGHTS = True
USE_DOWNHILL_SIMPLEX = True

class OptimalElution:
    def __init__(self, M, ecurve, fit_recs, rank, error=None, debug=False):
        x = ecurve.x

        pv_init = self.make_param_vector(fit_recs)
        self.fit_recs = fit_recs

        self.M = M
        E = error

        C = self.make_C_matrix(rank, pv_init, x)
        self.C_init = C
        self.P = M @ np.linalg.pinv(C)

        W_ = 1/E.copy()
        W_[np.logical_not(np.isfinite(W_))] = 0
        W_[W_<0] = 0

        if USE_FACTOED_WIEGHTS:
            U, s, VT = np.linalg.svd(W_)
            # self.scale = np.sqrt(np.linalg.norm(M))
            # self.scale = np.sqrt(s[0])
            nM = np.linalg.norm(M)
            scale = 1/np.sqrt(nM)
            print("scale=", scale)
            pw = U[:,0:1]*scale
            cw = VT[0:1,:]*scale
            W = pw @ cw
            print("norm(M)=", nM, "norm(W)=", np.linalg.norm(W))
        else:
            pw = np.sum(W_, axis=1)[:,np.newaxis]
            cw = np.sum(W_, axis=0)[np.newaxis,:]
            W = pw @ cw

        if False:
            from MatrixData import simple_plot_3d
            print('pw.shape=', pw.shape, 'ce.shape=', cw.shape, 'W.shape=', W.shape)
            plt.push()
            fig = plt.figure(figsize=(14, 6))
            ax1 = fig.add_subplot(121, projection='3d')
            ax2 = fig.add_subplot(122, projection='3d')
            ax1.set_title(r"Error Weights: $ \frac{1}{M_{error}} $", fontsize=32)
            ax2.set_title(r"Factored Weights: $ w_a @ w_e $", fontsize=32)
            simple_plot_3d(ax1, W_)
            simple_plot_3d(ax2, W)
            fig.tight_layout()
            plt.show()
            plt.pop()

        Mw = W*M
        h_indeces = np.arange(rank)*len(PARAMS)

        n = 0
        def objective_function(pv):
            nonlocal n
            # print('objective_function')
            C = self.make_C_matrix(rank, pv, x)
            Cw = C*cw
            Pw = Mw @ np.linalg.pinv(Cw)
            P = Pw/pw
            PC = P @ C
            D_ = W*(PC - M)
            v = np.linalg.norm(D_)
            if debug and n % 1000 == 0:
                print([n], 'v=', v)
                plt.push()
                gs = GridSpec(2,3)
                fig = plt.figure(figsize=(21,11))
                ax00 = fig.add_subplot(gs[0,0], projection='3d')
                ax01 = fig.add_subplot(gs[0,1], projection='3d')
                ax02 = fig.add_subplot(gs[0,2], projection='3d')
                ax10 = fig.add_subplot(gs[1,0], projection='3d')
                ax11 = fig.add_subplot(gs[1,1])
                ax12 = fig.add_subplot(gs[1,2])
                simple_plot_3d(ax00, M)
                simple_plot_3d(ax01, PC)
                simple_plot_3d(ax02, D_)
                simple_plot_3d(ax10, W)
                for k, p in enumerate(P.T):
                    ax11.plot(p)
                for k, c in enumerate(C):
                    ax12.plot(x, c)
                fig.tight_layout()
                plt.show()
                plt.pop()
            n += 1
            return v

        t0 = time()
        if USE_DOWNHILL_SIMPLEX:
            self.result = optimize.fmin(objective_function, pv_init,
                    maxiter=MAX_ITERATION_FMIN,
                    full_output=True)
        else:
            self.result = optimize.fmin_cg(objective_function, pv_init,
                    gtol=1e-7,
                    fprime=None,
                    maxiter=MAX_ITERATION_FMIN_CG,
                    full_output=True)

        print('I took ', time()-t0)

        self.x = x
        self.rank = rank

    def make_param_vector(self, fit_recs):
        param_list = []
        for rec in fit_recs:
            evaluator = rec.evaluator
            param_list.append([evaluator.get_param_value(p) for p in PARAMS])
        return np.array(param_list).flatten()

    def make_C_matrix(self, rank, pv, x):
        C_list = []
        pv_ = pv.reshape((rank, NUM_PARAMS))
        for h, mu, sigma, tau in pv_:
            C_list.append(self.model_func(x, h, mu, sigma, tau))
        C = np.array(C_list)
        return C

    def get_C_matrix(self):
        pv = self.result[0]
        return self.make_C_matrix(self.rank, pv, self.x)

class OptimalEgh(OptimalElution):
    def __init__(self, *args, **kwargs):
        self.model = EGH()
        self.model_func = egh
        self.model_foot = egh_x_from_height_ratio
        OptimalElution.__init__(self, *args, **kwargs)

class OptimalEmg(OptimalElution):
    def __init__(self, *args, **kwargs):
        self.model = EMG()
        self.model_func = emg
        self.model_foot = emg_x_from_height_ratio
        OptimalElution.__init__(self, *args, **kwargs)

def get_h_vector(fit_recs):
    h_vector = np.zeros(len(fit_recs))
    for k, rec in enumerate(fit_recs):
        evaluator = rec.evaluator
        h = evaluator.get_param_value(0)
        h_vector[k] = h

    max_h = np.max(h_vector)
    rank_guess = len(np.where(h_vector > max_h*0.05 )[0] )
    return h_vector, rank_guess

def select_resc(h_vector, rank, fit_recs):
    n = len(fit_recs) - rank
    pp = np.argpartition(h_vector, n)
    return [fit_recs[i] for i in sorted(pp[n:])]

def compute_optimal_elution(D, E, ecurve, model_type, legacy_info=None, rank=None, logger=None, debug=False):
    model = EGHA() if model_type else EMGA()

    if legacy_info is None:
        from CurveDecomposer import decompose
        fit_recs = decompose(ecurve, model=model)
        h_vector, rank_guess = get_h_vector(fit_recs)
    else:
        from DecompUtils import decompose_elution_better
        sd, mapper, corbase_info = legacy_info
        ret = decompose_elution_better(corbase_info, mapper, model,
                                    logger=logger)
        fit_recs = ret.opt_recs 
        h_vector, rank_guess = get_h_vector(fit_recs)

    if rank is None:
        rank = rank_guess

    fit_recs = select_resc(h_vector, rank, fit_recs)

    M = get_denoised_data(D, rank=rank)      # M

    if model_type:
        opimizer = OptimalEgh(M, ecurve, fit_recs, rank, error=E, debug=debug)
    else:
        opimizer = OptimalEmg(M, ecurve, fit_recs, rank, error=E, debug=debug)
    optC = opimizer.get_C_matrix()
    recs = opimizer.fit_recs

    return M, rank, opimizer.C_init, optC, recs



"""
    IntegerRatios.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np

def determine_integer_ratios(mws):
    ratios = mws/mws[-1]

    rounded_ratios = [round(v) for v in ratios]
    print("ratios=", ratios)
    print("rounded_ratios=", rounded_ratios)

    return np.array(rounded_ratios)



"""
module for Gaussian mixture model
"""
from .discrete_rv import DiscreteRV, assert_shape_equal
from .moments import hermite_transform_matrix, empirical_moment
import numpy as np

class ModelGM:
    """ class for 1-d Gaussian mixture model
    weights: ndarray of weights
    centers: ndarray of means
    sigma: standard deviations
    """
    def __init__(self, w=1, x=0, std=1):
        """ Initialize a GM model
        w: arrary of weights
        x: arrary of means
        std: scalar or array of standard deviations. If scalar, components share the same std.
        """

        self.weights = np.asarray(w)
        self.centers = np.asarray(x)
        assert_shape_equal(self.weights, self.centers)
        # np.testing.assert_array_equal(self.p.shape, self.mu.shape)
        
        if np.isscalar(std):
            self.sigma = std*np.ones(self.weights.shape)
        else:
            self.sigma = np.asarray(std)
        assert_shape_equal(self.weights, self.sigma)

    def __repr__(self):
        return "atom: %s\nwght: %s\nsigm: %s" % (self.centers, self.weights, self.sigma[0])

    def moments_gm(self, degree):
        """
        moments of GM model

        Args:
        degree: int
        highest degree k

        Returns:
        moments of Gaussian mixture model from degree 1 to k
        """
        mom = empirical_moment(self.centers/self.sigma, degree)
        transform = hermite_transform_matrix(degree)
        transform = (np.abs(transform[0]), np.abs(transform[1]))
        mom = np.dot(transform[0], mom) + transform[1]
        s_pow = empirical_moment(self.sigma, degree)
        mom = s_pow*mom
        return np.dot(mom, self.weights)

    def mean_rv(self):
        """
        discrete rv for the means
        """
        return DiscreteRV(self.weights, self.centers)

    def std_rv(self):
        """
        discrete rv for the sigmas
        """
        return DiscreteRV(self.weights, self.sigma)



def sample_gm(model, num):
    """
    n random samples from Gaussian mixture model
    """
    idx = np.random.choice(model.centers.shape[0], size=num, replace=True, p=model.weights)
    return model.centers[idx] + model.sigma[idx] * np.random.randn(num)



# coding: utf-8
"""

    TestUtils.py

    Copyright (c) 2016-2019, Masatsuyo Takahashi, KEK-PF
"""
import os
import shutil
import time
import numpy        as np
from subprocess     import call
import ctypes
from ctypes         import c_int, byref
from pyautogui      import screenshot, moveTo, dragTo, dragRel, click, mouseUp, mouseDown
import cv2
from PIL            import ImageGrab
from PIL            import Image    as ImagePIL

from molass_legacy.KekLib.BasicUtils     import mkdirs_with_retry, clear_dirs_with_retry
from TestTkUtils    import split_geometry, geometry_fix
from MachineTypes   import get_chassistype_name, get_display_resolution, get_monitors
# from PilatusImage   import PilatusImage

class _RECT( ctypes.Structure ):
    _fields_ =  [
                    ( 'left',    ctypes.c_int ),
                    ( 'top',     ctypes.c_int ),
                    ( 'right',   ctypes.c_int ),
                    ( 'bottom',  ctypes.c_int ),
                ]

def get_window_rect( hwid ):
    rect = _RECT( 0, 0, 0, 0 )
    ret = ctypes.windll.user32.GetWindowRect ( c_int( hwid ), byref(rect) )
    # print( ret )
    # print( rect.left, rect.top, rect.right, rect.bottom )
    return ( rect.left, rect.top, rect.right, rect.bottom )

def set_window_pos( hwid, after, x, y, cx, cy, flags ):
    ret = ctypes.windll.user32.SetWindowPos(
        c_int(hwid), c_int( after ), c_int( x ),  c_int( y ),  c_int( cx ),  c_int( cy ), c_int( flags ) )
    # print( 'ret=', ret )
    if ret == 0:
        print( ctypes.FormatError() )
    return ret

def single_screenshot( hwid, file ):
    l, t, r, b = get_window_rect( hwid )
    im = ImageGrab.grab( bbox=( l, t, r, b ) )
    im.save( file )

def open_explorer( folder, at=None ):
    call( 'explorer %s' % ( folder ), shell=True )

    if at == None: return None

    assert( len( at ) == 2 )

    name = os.path.split( folder )[-1]

    retry   = 0
    hwid    = 0
    while( retry < 3 and hwid == 0 ):
        time.sleep( 1 )
        hwid = ctypes.windll.user32.FindWindowW( None, name )
        retry += 1

    if hwid == 0:
        print( "Can't find window titled '%s'" % ( name ) )
        return None

    print( "Found window titled '%s'" % ( name ) )

    x, y, r, b = get_window_rect( hwid )
    print(  x, y, r, b )

    if get_chassistype_name() == 'Desktop':
        set_window_pos( hwid, 0, at[0], at[1], 800, 400, 0 )
    else:
        # notebook では set_window_pos すると、後続の drag のタイミングが
        # よくないようなので、drag で位置調整する。
        x_      = ( x + r ) / 2
        y_      = y + 20

        rel_x   = at[0] - x
        rel_y   = at[1] - y

        moveTo( x_, y_ )
        dragRel( rel_x, rel_y, 0.5 )

    return hwid

def close_explorer( hwid ):
    _, t, r, _ = get_window_rect( hwid )
    x, y = r - 30, t + 10
    moveTo( x, y )
    click()

def dnd_from_to( from_folder, dnd_obj_image, root, to_widget, screenshot_only=False ):
    display_width, _    = get_display_resolution()
    monitors            = get_monitors()
    print( 'monitors=', monitors )
    if len( monitors ) == 1:
        display_left_position = 0
    else:
        display_left_position = display_width + 500
    print( 'display_left_position=', display_left_position )

    geometry_fix( root, display_left_position, 0 )

    print( 'from_folder=', from_folder )
    hwid = open_explorer( from_folder, at=( display_left_position + 200, 230 ) )
    time.sleep( 1 )
    temp_dir = 'dnd_from_to_temp'
    clear_dirs_with_retry( [ temp_dir ] )
    screenshot_im_file = temp_dir + '/folder_screenshot.png'

    # screenshot( screenshot_im_file )
    single_screenshot( hwid, screenshot_im_file )

    fx, fy, fr, fb = get_window_rect( hwid )
    print( 'fx, fy, fr, fb=', fx, fy, fr, fb )

    if screenshot_only:
        return

    from_screen = cv2.imread( screenshot_im_file )
    obj_image   = cv2.imread( dnd_obj_image )
    ( objHeight, objWidth ) = obj_image.shape[:2]

    result = cv2.matchTemplate( from_screen, obj_image, cv2.TM_CCOEFF )
    (_, _, minLoc, maxLoc) = cv2.minMaxLoc(result)
    print( 'result=', minLoc, maxLoc )
    topLeft = maxLoc
    objX = fx + topLeft[0] + objWidth/2
    objY = fy + topLeft[1] + objHeight/2

    W, H, X, Y = split_geometry( root.geometry() )
    print( "root geometry", W, H, X, Y )
    w, h, x, y = split_geometry( to_widget.winfo_geometry() )
    print( "to_widget geometry", w, h, x, y )

    to_widget_ok = False

    def dummy():
        pass

    while( not to_widget_ok ):
        moveTo( objX, objY )
        time.sleep( 0.5 )

        to_widget.focus_force()

        x_ = X + x + w/2
        y_ = Y + y + h/2 + 30
        dragTo( x_, y_, 0.5 )
        mouseDown( x_, y_ )
        mouseUp( x_, y_ )
        time.sleep( 0.5 )
        for k in range(3):
            root.update()
        value = to_widget.get()
        print( value )
        if value.find( '<' ) < 0:
            to_widget_ok = True

    close_explorer( hwid )
    shutil.rmtree( temp_dir )

def OldPilatusImage( path, original_image=None ):
    assert( original_image is not None )
    from PilatusImage import PilatusImage
    im = ImagePIL.open( path )
    im_array = np.array( im )
    return PilatusImage( im_array, original_image=original_image )

def max_diff(a, b):
    """
    Elegant way to perform tuple arithmetic
    https://stackoverflow.com/questions/17418108/elegant-way-to-perform-tuple-arithmetic
    """
    try:
        return np.max(np.subtract(a, b))
    except Exception as exc:
        raise ValueError( str(exc) + '\nvalues=' + '\na=' + str(a) + '\nb=' + str(b) )

def max_diff_assert_lt(a, b, c):
    try:
        assert max_diff(a, b) < c
    except Exception as exc:
        raise AssertionError( str(exc) + '\nvalues=' + '\na=' + str(a) + '\nb=' + str(b) )

def get_filesize(path):
    """
    How to check file size in python?
    https://stackoverflow.com/questions/2104080/how-to-check-file-size-in-python
    """
    statinfo = os.stat(path)
    return statinfo.st_size



# coding: utf-8
"""
    SecTheory.Sepsey2014.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy import integrate

def rho(rg, rp):
    return 0 if rg > rp else rg/rp

def np_(nperm, me, rho_):
    return nperm*(1 - rho_)**me

def tp_(tperm, mp, rho_):
    return tperm*(1 - rho_)**mp

def lognormal_pore_cf_iterate(w, rp0, sigma, nperm, me, tperm, mp, rg):
    # 2014, A. Sepsey. Eq.(21)
    z = 0
    denom_sigma2 = 2*sigma**2
    log_rp0 = np.log(rp0)
    for r in range(10, 200):
        if r < rg:
            continue

        rho_ = rg/r
        np_ = nperm*(1 - rho_)**me
        tp_ = tperm*(1 - rho_)**mp
        z += np_/r * np.exp(- (np.log(r) - log_rp0)**2/denom_sigma2) * (1/(1 - 1j*w*tp_) - 1)
    z /= np.sqrt(2*np.pi)*sigma
    return np.exp(z)

def lognormal_pore_cf(w, rp0, sigma, nperm, me, tperm, mp, rg):
    # 2014, A. Sepsey. Eq.(21)
    denom_sigma2 = 2*sigma**2
    log_rp0 = np.log(rp0)
    def integrand(r):
        rho_ = rg/r
        np_ = nperm*(1 - rho_)**me
        tp_ = tperm*(1 - rho_)**mp
        return np_/r * np.exp(- (np.log(r) - log_rp0)**2/denom_sigma2) * (1/(1 - 1j*w*tp_) - 1)
    val, err = integrate.quad_vec(integrand, rg, 100)
    return np.exp(np.sqrt(2*np.pi)*sigma*val)



"""
Computing the 3D Radial Distribution Function from Particle Positions: An Advanced Analytic Approach
Bernd A. F. Kopera and Markus Retsch
Analytical Chemistry 2018 90 (23), 13909-13914
DOI: 10.1021/acs.analchem.8b03157
"""

from numpy import sqrt, pi, zeros, ones, arctan
# from random import random, seed
# from time import time
# seed(time())

"""
This function calculates the pair correlation functrion g(2)(r)
for a dense set of particles in a rectanglar box. The partricle
coordinates are supplied as [[x, y, z],...] lists.
All particles are used as central aprticles.
The spherical shell might extend beyond the known region.
An empty surrounding is assumed.
"""

"""***********************************************************"""

def SphereCutVol(Rs, A, B):

    Root = sqrt(Rs**2-A**2-B**2)

    Vcut = 1/6*Rs**3*(pi - 2*arctan(A*B/(Rs*Root)))
    Vcut += 1/2 * (arctan(A/Root) - pi/2)*(Rs**2*B-1/3*B**3)
    Vcut += 1/2 * (arctan(B/Root) - pi/2)*(Rs**2*A-1/3*A**3)
    Vcut += 1/3 * A * B * Root

    return Vcut

"""***********************************************************"""

def OctVolume(Rs, xb, yb, zb):

    # if all boundaries are fully in the octant
    if xb**2 + yb**2 + zb **2 < Rs**2:

        return xb * yb * zb

    # if no boundary intersects we start with
    VOctant = 1/8 * 4/3 * pi * Rs**3

    # remove the spherical caps
    for B in [xb , yb, zb]:

        if B < Rs:

            VOctant -= pi/4*(2/3*Rs**3-B*Rs**2+1/3*B**3)

    # add the intersections of the caps
    for (a, b) in [(xb, yb), (xb, zb), (yb ,zb)]:

        if a**2 + b**2 < Rs**2:

            VOctant += SphereCutVol(Rs, a, b)

    return VOctant

"""**********************************************************"""

def SphereVolume(Rs, BoxBounds):

    [Xmin, Xmax, Ymin, Ymax, Zmin, Zmax] = BoxBounds

    VSphere = 0

    # abs() mirrors the boundaries into the first octant
    for xb in [Xmin, Xmax]:
        for yb in [Ymin ,Ymax]:
            for zb in [Zmin, Zmax]:

                VSphere += OctVolume(Rs, abs(xb), abs(yb), abs(zb))

    return VSphere

"""***********************************************************"""

def ShellVolume(Rmin, Rmax, BoxBounds):

    # check for negative Rmin values
    Rmin = max([Rmin, 0])

    InnerShell = SphereVolume(Rmin, BoxBounds)
    OuterShell = SphereVolume(Rmax, BoxBounds)

    Volume = OuterShell - InnerShell

##  if Volume <= 0:
##      print("Volume = " + str(Volume))
##      print("Rmin = " + str(Rmin))
##      print("Rmax = " + str(Rmax))
##      print("BoxBounds: " + str(BoxBounds))

    return Volume

"""************************************************************"""

def RDF_AnalyticNorm(Particles, r, dr):

    # Gr averaged over all particles
    Global_Gr = zeros(len(r))

    # Keep track of the use full shell volumes
    NonEmptyShells = zeros(len(r))

    # maximal radial distance
    MaxDist = r[-1] + dr/2

    # Box boundaries
    XList = [Particles[k][0] for k in range(0, len(Particles), 1)]
    YList = [Particles[k][1] for k in range(0, len(Particles), 1)]
    ZList = [Particles[k][2] for k in range(0, len(Particles), 1)]

    BoxBounds =[min(XList), max(XList),
                min(YList), max(YList),
                min(ZList), max(ZList)]

    # box size
    Lx = BoxBounds[1] - BoxBounds [0]
    Ly = BoxBounds[3] - BoxBounds [2]
    Lz = BoxBounds[5] - BoxBounds [4]

    print("Lx = " + str(Lx))
    print("Ly = " + str(Ly))
    print("Lz = " + str(Lz))

    MeanDensity = len(Particles) / (Lx * Ly * Lz)

    # use every particle as the center once
    for CentralP in range (0, len(Particles), 1):

        # local Gr around the current particle
        Local_Gr = zeros(len(r))

        # look at every other particle at most MaxDist away:
        for Neighbour in range(0, len(Particles), 1):

            if CentralP != Neighbour :

                # calc the distance to the neighbour
                dx = Particles[CentralP][0] - Particles[Neighbour][0]
                dy = Particles[CentralP][1] - Particles[Neighbour][1]
                dz = Particles[CentralP][2] - Particles[Neighbour][2]

                d = sqrt(dx**2 + dy**2 + dz**2)

                # what bins is the particle in?
                IdxList = [k for k in range(0, len(r), 1) if abs (r[k] - d) <= dr/ 2]

                # add one to every bin the particle is in
                for Pos in IdxList:

                    # count the particle
                    Local_Gr[Pos] += 1

        # shift the center of box cosy
        LocalBox = [BoxBounds[0] - Particles[CentralP][0],
                    BoxBounds[1] - Particles[CentralP][0],
                    BoxBounds[2] - Particles[CentralP][1],
                    BoxBounds[3] - Particles[CentralP][1],
                    BoxBounds[4] - Particles[CentralP][2],
                    BoxBounds[5] - Particles[CentralP][2]]

        # normalize with the shell volume
        for RIdx in range(0, len(r), 1):

            SVolume = ShellVolume(r[RIdx]-dr/2, r[RIdx]+dr/2, LocalBox)

            if SVolume > 0.0:

                Local_Gr[RIdx] /= SVolume
                NonEmptyShells[RIdx] += 1

        # normalize by the mean particle density
        Local_Gr = Local_Gr / MeanDensity

        # save in the global gr for the average over particles
        Global_Gr = Global_Gr + Local_Gr

        print("Finished Particle " + str(CentralP) + " of " + str(len(Particles)))

    # final normalization considering the non empty shell volumes

    for k in range(0, len(Global_Gr), 1):

        if NonEmptyShells[k] != 0:

            Global_Gr[k] /= NonEmptyShells[k]

        else:
            print("All Shells at R = " + str(r[k]) + " are Empty!" )

    return Global_Gr

"""************************************************************"""



"""

    Peaks.EghSupples.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF

"""
import numpy as np
from molass_legacy.Peaks.ElutionModels import egh

"""

See also, 2001, Kevin Lan, James W. Jorgenson

from sympy import symbols
from sympy.solvers import solve
s, t, a, A, B = symbols("s, t, a, A, B")
solve((s**2 + B*A/(2*a), t + (B-A)/a), A, B)
[(a*t/2 - sqrt(-a*(-a*t**2 + 8*s**2))/2,
  -a*t/2 - sqrt(-a*(-a*t**2 + 8*s**2))/2),
 (a*t/2 + sqrt(-a*(-a*t**2 + 8*s**2))/2,
  -a*t/2 + sqrt(-a*(-a*t**2 + 8*s**2))/2)]

"""
def compute_AB(a, s, t):
    # a: np.log(alpha)
    A = a*t/2 + np.sqrt(-a*(-a*t**2 + 8*s**2))/2
    B = -a*t/2 + np.sqrt(-a*(-a*t**2 + 8*s**2))/2
    return A, B

def d_egh(x, H, tR, sigma, tau):
    x_  = x - tR
    s2  = 2 * sigma**2
    z   = s2 + tau*x_
    z_neg   = z <= 0
    z_pos   = z > 0

    zero_part = np.zeros( len(x) )[z_neg]

    # t*(x - m)**2/(2*s**2 + t*(x - m))**2 - 2*(x - m)/(2*s**2 + t*(x - m))
    xp = x_[z_pos]
    denom = s2 + tau*xp 
    posi_part = (tau*xp **2/denom**2 - 2*xp/denom) * egh(x[z_pos], H, tR, sigma, tau)

    if tau > 0:
        parts = [ zero_part, posi_part ]
    else:
        parts = [ posi_part, zero_part ]

    return np.hstack( parts )



"""
    SecTheory.MonoporeSimulation.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
from matplotlib.gridspec import GridSpec

def demo(use_cf=True):
    import molass_legacy.KekLib.DebugPlot as plt
    if use_cf:
        from SecTheory.SecCF import gec_monopore_phi
        from SecTheory.SecPDF import FftInvPdf
        monopore_pdf = FftInvPdf(gec_monopore_phi)
    else:
        from SecTheory.BasicModels import robust_single_pore_pdf
        monopore_pdf = robust_single_pore_pdf

    x = np.arange(300)

    me = 0.1
    mp = 1.5
    rp = 76
    rg = 35
    rho = rg/rp
    N = 171
    T = 0.63
    t0 = 100

    rg_array = np.array([60, 50, 40, 30, 20, 10, 5])

    y_list = []
    for rg_ in rg_array:
        rho_ = rg_/rp
        np_ = N*(1 - rho_)**me
        tp_ = T*(1 - rho_)**mp
        if use_cf:
            y = monopore_pdf(x, np_, tp_, t0)
        else:
            y = monopore_pdf(x - t0, np_, tp_)
        y_list.append(y)

    with plt.Dp():
        fig, ax = plt.subplots()

        for y, rg_ in zip(y_list, rg_array):
            ax.plot(x, y, label="$R_g=%.3g$" % rg_)

        ax.legend()
        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    demo()



"""
    DenssFitData.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF
"""
import os, sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Slider, Button, RadioButtons, TextBox, CheckButtons #, RangeSlider
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ScrolledText
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.StdoutRedirector import StdoutRedirector

class DenssFitDataDialog(Dialog):
    def __init__(self, parent, sasrec, work_info, infile_name, out_folder):
        self.sasrec = sasrec
        self.work_info = work_info
        self.infile_name = infile_name
        self.out_folder = out_folder
        self.applied = False
        Dialog.__init__(self, parent, "Figure in denss_fit_data.py", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        set_icon( self )

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.log_text = ScrolledText(body_frame, height=8)
        self.log_text.pack(fill=Tk.X)
        self.redirexctor = StdoutRedirector(self.log_text)

        fig = self.create_figure(cframe)
        name, ext = os.path.split(self.infile_name)
        output = os.path.join(self.out_folder, name + '_fit')
        if not os.path.exists(self.out_folder):
            os.makedirs(self.out_folder)
        self.buttons = fit_data_plot(self.sasrec, self.work_info, fig, output, tkinter_gui=True)
        self.mpl_canvas.draw()

    def create_figure(self, cframe):
        self.fig = fig = plt.figure(figsize=(14,7))
        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        return fig

    def apply(self):
        self.applied = True

    def get_sasrec(self):
        global sasrec
        return sasrec

def fit_data_plot(init_sasrec, work_info, fig, output, tkinter_gui=False):
    # from .saxstats import saxstats as saxs
    import molass_legacy.DENSS.denss as denss
    global sasrec

    args = work_info.args
    alpha = work_info.alpha
    n1 = work_info.n1
    n2 = work_info.n2
    Iq_orig = work_info.Iq_orig
    r = args.r
    nes = args.nes

    sasrec = init_sasrec
    qc = sasrec.qc
    D = sasrec.D            # has been initialized by saxs.estimate_dmax(Iq) in DenssUtils.fit_data_impl
    Iq = work_info.Iq
    Icerr = sasrec.Icerr

### DENSS.bin.denss.fit_data.py copy & modify BEIGN ###
    #set a maximum alpha range, so when users click in the slider
    #it at least does something reasonable, rather than either nothing
    #significant, or so huge it becomes difficult to find the right value
    if args.max_alpha is None:
        if alpha == 0.0:
            max_alpha = 2.0
        else:
            max_alpha = 2*alpha

    def store_parameters_as_string(event=None):
        param_str = ("Parameter Values:\n"
        "Chi2  = {chi2:.5e}\n"
        "Dmax  = {dmax:.5e}\n"
        "alpha = {alpha:.5e}\n"
        "I(0)  = {I0:.5e} +- {I0err:.5e}\n"
        "Rg    = {rg:.5e} +- {rgerr:.5e}\n"
        "r_avg = {r:.5e} +- {rerr:.5e}\n"
        "Vp    = {Vp:.5e} +- {Vperr:.5e}\n"
        "MW_Vp = {mwVp:.5e} +- {mwVperr:.5e}\n"
        "Vc    = {Vc:.5e} +- {Vcerr:.5e}\n"
        "MW_Vc = {mwVc:.5e} +- {mwVcerr:.5e}\n"
        "Lc    = {lc:.5e} +- {lcerr:.5e}\n"
        ).format(chi2=sasrec.chi2,
            dmax=sasrec.D,alpha=sasrec.alpha,
            I0=sasrec.I0,I0err=sasrec.I0err,
            rg=sasrec.rg,rgerr=sasrec.rgerr,
            r=sasrec.avgr,rerr=sasrec.avgrerr,
            Vp=sasrec.Vp,Vperr=sasrec.Vperr,
            mwVp=sasrec.mwVp,mwVperr=sasrec.mwVperr,
            Vc=sasrec.Vc,Vcerr=sasrec.Vcerr,
            mwVc=sasrec.mwVc,mwVcerr=sasrec.mwVcerr,
            lc=sasrec.lc,lcerr=sasrec.lcerr)
        return param_str

    def print_values(event=None):
        print("---------------------------------")
        param_str = store_parameters_as_string()
        print(param_str)

    def save_file(event=None):
        param_str = store_parameters_as_string()
        #add column headers to param_str for output
        param_str += 'q, I, error, fit ; chi2 = %.3f'%sasrec.chi2
        #quick, interpolate the raw data, sasrec.I, to the new qc values, but be sure to
        #put zeros in for the q values not measured
        Iinterp = np.interp(sasrec.qc, sasrec.q_data, sasrec.I_data, left=0.0, right=0.0)
        Ierrinterp = np.interp(sasrec.qc, sasrec.q_data, sasrec.Ierr_data)
        np.savetxt(output+'.fit', np.vstack((sasrec.qc, Iinterp, Ierrinterp, sasrec.Ic)).T,delimiter=' ',fmt='%.5e',header=param_str)
        np.savetxt(output+'_pr.dat', np.vstack((sasrec.r, sasrec.P, sasrec.Perr)).T,delimiter=' ',fmt='%.5e')
        print("%s and %s files saved" % (output+".fit",output+"_pr.dat"))
        if args.write_shannon:
            np.savetxt(output+'_Shannon.dat', np.vstack((sasrec.qn, sasrec.In, sasrec.Inerr)).T,delimiter=' ',fmt='%.5e',header=param_str)

    if tkinter_gui:
        # using FigureCanvasTkAgg
        import matplotlib.pyplot as plt
        from matplotlib.widgets import Slider, Button, RadioButtons, TextBox, CheckButtons #, RangeSlider
    elif args.plot:
        #if plotting is enabled, try to import matplotlib
        #if import fails, set plotting to false
        #first try using Qt5Agg backend, then TkAgg
        try:
            import matplotlib
            matplotlib.use('Qt5Agg')
            import matplotlib.pyplot as plt
            from matplotlib.widgets import Slider, Button, RadioButtons, TextBox, CheckButtons #, RangeSlider
            fig = plt.figure(0, figsize=(12,6))
            qtsuccess = True
        except ImportError:
            qtsuccess = False
        if not qtsuccess:
            print("Using TkAgg")
            try:
                import matplotlib
                matplotlib.use('TkAgg')
                import matplotlib.pyplot as plt
                from matplotlib.widgets import Slider, Button, RadioButtons, TextBox, CheckButtons #, RangeSlider
                fig = plt.figure(0, figsize=(12,6))
            except Exception as e:
                print("Matplotlib failed to create figure. GUI mode disabled.")
                print("Error: %s"%e)
                args.plot = False
        else:
            print("Using QtAgg")

    if args.plot:
        try:
            #this command changed in matplotlib v3.6
            fig.canvas.manager.set_window_title(output)
        except:
            try:
                fig.canvas.set_window_title(output)
            except:
                pass
        fig.suptitle(output)
        axI = plt.subplot2grid((3,2), (0,0),rowspan=2)
        axR = plt.subplot2grid((3,2), (2,0),sharex=axI)
        axP = plt.subplot2grid((3,2), (0,1),rowspan=3)
        fig.subplots_adjust(left=0.068, bottom=0.25, right=0.98, top=0.95)

        #add a plot of untouched light gray data for reference for the user
        I_l0, = axI.plot(Iq_orig[:,0], Iq_orig[:,1], '.', c='0.8', ms=3)
        I_l1, = axI.plot(sasrec.q_data, sasrec.I_data, 'k.', ms=3, label='test')
        I_l2, = axI.plot(sasrec.qc, sasrec.Ic, 'r-', lw=2)
        I_l3, = axI.plot(sasrec.qn, sasrec.In, 'bo', mec='b', mfc='none', mew=2)
        chi2_text = axI.text(0.7,0.9,r"$\chi^2$ = %.3e"%sasrec.chi2,transform=axI.transAxes,fontsize='large')
        if args.log: axI.semilogy()
        axI.set_ylabel('I(q)')
        axI.set_xlabel('q')

        #make plots for each Bn (should we also do each Sn for P(r) plot?)
        # I_Bn = []
        # nn = range(5,20) #range(20) #idx Bns to show
        # for i in nn:
        #     I_Bn.append(axI.plot(sasrec.q, 2*sasrec.In[i]*sasrec.B[i], label='B_%d'%i))


        #residuals
        #first, just ensure that we're comparing similar q ranges, so
        #interpolate from qc to q_data to enable subtraction, since there's
        #q values in qc that are not in q_data, and possibly vice versa
        Icinterp = np.interp(sasrec.q_data, sasrec.qc, sasrec.Ic)
        res = (sasrec.I_data - Icinterp)/sasrec.Ierr_data
        #in case qc were fewer points than the data, for whatever reason,
        #only grab the points up to qc.max
        ridx = np.where((sasrec.q_data<sasrec.qc.max()))
        absolute_maximum_q = np.max([sasrec.qc.max(),sasrec.q.max(),Iq_orig[:,0].max()])
        Ires_l0, = axR.plot([0,absolute_maximum_q], [0,0], 'k--')
        Ires_l1, = axR.plot(sasrec.q_data[ridx], res[ridx], 'r.', ms=3)
        axR.set_ylabel('Residuals')
        axR.set_xlabel('q')

        P_l1, = axP.plot(sasrec.r*100, sasrec.r*0, 'k--')
        P_l2, = axP.plot(sasrec.r, sasrec.P, 'b-', lw=2)
        axP.set_ylabel('P(r)')
        axP.set_xlabel('r')

        #plot Sns on P(r)
        # P_Sn = []
        # for i in nn:
        #     P_Sn.append(axP.plot(sasrec.r, sasrec.In[i]*sasrec.S[i], label='S_%d'%i))
        # mm = nn #range(0,20)
        # P_Sn[-1] = axP.plot(sasrec.r, np.sum(sasrec.In[mm,None]*sasrec.S[mm],axis=0),'k-',lw=4)

        #axI.set_xlim([0,1.1*np.max(sasrec.q)])
        #axR.set_xlim([0,1.1*np.max(sasrec.q)])
        axI.set_xlim([0,Iq_orig[-1,0]])
        axR.set_xlim([0,Iq_orig[-1,0]])
        axP.set_xlim([0,1.1*np.max(sasrec.r)])
        axI.set_ylim([0.25*np.min(sasrec.Ic[sasrec.qc<Iq_orig[-1,0]]),2*np.max(sasrec.Ic[sasrec.qc<Iq_orig[-1,0]])])
        #axR.set_ylim([0,Iq_orig[-1,0]])
        #axP.set_ylim([0,1.1*np.max(sasrec.r)])
        #the "q" axis label is a little low, so let's raise it up a bit
        axR.xaxis.labelpad = -10

        axcolor = 'lightgoldenrodyellow'
        #axn1n2 = plt.axes([0.05, 0.175, 0.4, 0.03], facecolor=axcolor)
        axdmax = plt.axes([0.05, 0.125, 0.4, 0.03], facecolor=axcolor)
        axalpha = plt.axes([0.05, 0.075, 0.4, 0.03], facecolor=axcolor)
        #axnes = plt.axes([0.05, 0.025, 0.4, 0.03], facecolor=axcolor)

        axI0 = plt.figtext(.57, .125,   r"$I(0)$  = %.2e $\pm$ %.2e"%(sasrec.I0,sasrec.I0err),family='monospace')
        axrg = plt.figtext(.57, .075,   r"$R_g$   = %.2e $\pm$ %.2e"%(sasrec.rg,sasrec.rgerr),family='monospace')
        axrav = plt.figtext(.57, .025,  r"$\overline{r}$    = %.2e $\pm$ %.2e"%(sasrec.avgr,sasrec.avgrerr),family='monospace')
        axVp = plt.figtext(.77, .125,   r"$V_p$ = %.2e $\pm$ %.2e"%(sasrec.Vp,sasrec.Vperr),family='monospace')
        axVc = plt.figtext(.77, .075,   r"$V_c$ = %.2e $\pm$ %.2e"%(sasrec.Vc,sasrec.Vcerr),family='monospace')
        axlc = plt.figtext(.77, .025,   r"$\ell_c$ = %.2e $\pm$ %.2e"%(sasrec.lc,sasrec.lcerr),family='monospace')
        #axVpmw = plt.figtext(.55, .075, "Vp MW = %.2e $\pm$ %.2e"%(sasrec.mwVp,sasrec.mwVperr),family='monospace')
        #axVcmw = plt.figtext(.55, .025, "Vc MW = %.2e $\pm$ %.2e"%(sasrec.mwVc,sasrec.mwVcerr),family='monospace')

        #RangeSlider is for very new versions of matplotlib, so for now ignore it
        #sn1n2 = RangeSlider(axn1n2, 'n1n2', 0, Iq_orig.shape[0], valinit=(n1, n2))
        #sn1n2.valtext.set_visible(False)

        sdmax = Slider(axdmax, 'Dmax', 0.0, args.max_dmax, valinit=D)
        sdmax.valtext.set_visible(False)
        # set up ticks marks on the slider to denote the change in interaction
        axdmax.set_xticks([0.9 * sdmax.valmax, 0.1 * sdmax.valmax])
        #axdmax.xaxis.tick_top()
        axdmax.tick_params(labelbottom=False)

        salpha = Slider(axalpha, r'$\alpha$', 0.0, max_alpha, valinit=alpha)
        salpha.valtext.set_visible(False)

        dmax = D
        n1 = str(n1)
        n2 = str(n2)
        nRemove = ""

        def analyze(dmax,alpha,n1,n2,nRemove,extrapolate):
            global sasrec
            points = np.arange(Iq_orig.shape[0])
            mask = np.ones(len(points),dtype=bool)
            mask[points<=n1] = False
            mask[points>=n2] = False
            #parse nRemove string, allow n point or n-m ranges
            ranges = nRemove.split(",")
            for rangei in ranges:
                if "-" in rangei:
                    rangei_a, rangei_b = [int(j) for j in rangei.split("-")]
                    mask[(points>=rangei_a)&(points<=rangei_b)] = False
                elif rangei == "":
                    pass
                else:
                    try:
                        mask[int(rangei)] = False
                    except:
                        print("Invalid Remove Range")
            qc = denss.create_lowq(q=Iq_orig[:,0])
            sasrec = denss.Sasrec(Iq_orig[mask], dmax, qc=qc, r=r, nr=args.nr, alpha=alpha, ne=nes, extrapolate=extrapolate)
            sasrec.estimate_Vp_etal()
            res = (sasrec.I_data - sasrec.Ic_qe)/sasrec.Ierr_data
            ridx = np.where((sasrec.q_data<sasrec.qc.max()))
            I_l1.set_data(sasrec.q_data, sasrec.I_data)
            I_l2.set_data(sasrec.qc, sasrec.Ic)
            I_l3.set_data(sasrec.qn, sasrec.In)
            chi2_text.set_text(r"$\chi^2$ = %.3e"%sasrec.chi2)
            Ires_l1.set_data(sasrec.q_data, res)
            P_l2.set_data(sasrec.r, sasrec.P)
            axI0.set_text(r"$I(0)$  = %.2e $\pm$ %.2e"%(sasrec.I0,sasrec.I0err))
            axrg.set_text(r"$R_g$   = %.2e $\pm$ %.2e"%(sasrec.rg,sasrec.rgerr))
            axrav.set_text(r"$\overline{r}$    = %.2e $\pm$ %.2e"%(sasrec.avgr,sasrec.avgrerr))
            axVp.set_text(r"$V_p$ = %.2e $\pm$ %.2e"%(sasrec.Vp,sasrec.Vperr))
            axVc.set_text(r"$V_c$ = %.2e $\pm$ %.2e"%(sasrec.Vc,sasrec.Vcerr))
            axlc.set_text(r"$\ell_c$ = %.2e $\pm$ %.2e"%(sasrec.lc,sasrec.lcerr))
            #axVpmw.set_text("Vp MW = %.2e $\pm$ %.2e"%(sasrec.mwVp,sasrec.mwVperr))
            #axVcmw.set_text("Vc MW = %.2e $\pm$ %.2e"%(sasrec.mwVc,sasrec.mwVcerr))
            # j = 0
            # for i in nn:
            #     I_Bn[j][0].set_data(sasrec.q, 2*sasrec.In[i]*sasrec.B[i])
            #     j+=1
            # j = 0
            # for i in nn:
            #     P_Sn[j][0].set_data(sasrec.r, sasrec.In[i]*sasrec.S[i])
            #     j+=1
            # P_Sn[-1][0].set_data(sasrec.r, np.sum(sasrec.In[mm,None]*sasrec.S[mm],axis=0))


        def n1_submit(text):
            dmax = sdmax.val
            alpha = salpha.val
            n1 = int(text)
            n2 = int(n2_box.text)
            nRemove = str(nRemove_box.text)
            extrapolate = extrapolate_check.get_status()[0]
            analyze(dmax,alpha,n1,n2,nRemove,extrapolate)
            fig.canvas.draw_idle()

        def n2_submit(text):
            dmax = sdmax.val
            alpha = salpha.val
            n1 = int(n1_box.text)
            n2 = int(text)
            nRemove = str(nRemove_box.text)
            extrapolate = extrapolate_check.get_status()[0]
            analyze(dmax,alpha,n1,n2,nRemove,extrapolate)
            fig.canvas.draw_idle()

        def nRemove_submit(text):
            dmax = sdmax.val
            alpha = salpha.val
            n1 = int(n1_box.text)
            n2 = int(n2_box.text)
            nRemove = str(text)
            extrapolate = extrapolate_check.get_status()[0]
            analyze(dmax,alpha,n1,n2,nRemove,extrapolate)
            fig.canvas.draw_idle()

        def extrapolate_submit(text):
            dmax = sdmax.val
            alpha = salpha.val
            n1 = int(n1_box.text)
            n2 = int(n2_box.text)
            nRemove = str(nRemove_box.text)
            extrapolate = extrapolate_check.get_status()[0]
            analyze(dmax,alpha,n1,n2,nRemove,extrapolate)
            fig.canvas.draw_idle()

        def D_submit(text):
            dmax = float(text)
            alpha = salpha.val
            n1 = int(n1_box.text)
            n2 = int(n2_box.text)
            nRemove = str(nRemove_box.text)
            extrapolate = extrapolate_check.get_status()[0]
            analyze(dmax,alpha,n1,n2,nRemove,extrapolate)
            # this updates the slider value based on text box value
            sdmax.set_val(dmax)
            if (dmax > 0.9 * sdmax.valmax) or (dmax < 0.1 * sdmax.valmax):
                sdmax.valmax = 2 * dmax
                sdmax.ax.set_xlim(sdmax.valmin, sdmax.valmax)
                axdmax.set_xticks([0.9 * sdmax.valmax, 0.1 * sdmax.valmax])
            fig.canvas.draw_idle()

        def A_submit(text):
            dmax = sdmax.val
            alpha = float(text)
            n1 = int(n1_box.text)
            n2 = int(n2_box.text)
            nRemove = str(nRemove_box.text)
            extrapolate = extrapolate_check.get_status()[0]
            analyze(dmax,alpha,n1,n2,nRemove,extrapolate)
            # this updates the slider value based on text box value
            salpha.set_val(alpha)
            # partions alpha slider
            if (alpha > 0.9 * salpha.valmax) or (alpha < 0.1 * salpha.valmax):
                salpha.valmax = 2 * alpha
                # alpha starting at zero makes initial adjustment additive not multiplicative
                if alpha != 0:
                    salpha.ax.set_xlim(salpha.valmin, salpha.valmax)
                elif alpha == 0:
                    salpha.valmax = alpha + 10
                    salpha.valmin = 0.0
                    salpha.ax.set_xlim(salpha.valmin, salpha.valmax)
            fig.canvas.draw_idle()

        def update(val):
            dmax = sdmax.val
            alpha = salpha.val
            n1 = int(n1_box.text)
            n2 = int(n2_box.text)
            nRemove = str(nRemove_box.text)
            extrapolate = extrapolate_check.get_status()[0]
            #print(extrapolate)
            analyze(dmax,alpha,n1,n2,nRemove,extrapolate)
            # partitions the slider, so clicking in the upper and lower range scale valmax
            if (dmax > 0.9 * sdmax.valmax) or (dmax < 0.1 * sdmax.valmax):
                sdmax.valmax = 2 * dmax
                sdmax.ax.set_xlim(sdmax.valmin, sdmax.valmax)
                axdmax.set_xticks([0.9 * sdmax.valmax, 0.1 * sdmax.valmax])
            # partitions slider as well
            if (alpha > 0.9 * salpha.valmax) or (alpha < 0.1 * salpha.valmax):
                salpha.valmax = 2 * alpha
                # alpha starting at zero makes initial adjustment additive not multiplicative
                if alpha != 0:
                    salpha.ax.set_xlim(salpha.valmin, salpha.valmax)
                elif alpha == 0:
                    salpha.valmax = alpha + 10
                    salpha.valmin = 0.0
                    salpha.ax.set_xlim(salpha.valmin, salpha.valmax)

            Dmax_box.set_val("%.4e"%dmax)
            Alpha_box.set_val("%.4e"%alpha)

            fig.canvas.draw_idle()

        # making a text entry for dmax that allows for user input
        Dvalue = "{dmax:.4e}".format(dmax=dmax)
        axIntDmax = plt.axes([0.45, 0.125, 0.08, 0.03])
        Dmax_box = TextBox(axIntDmax, '', initial=Dvalue)
        Dmax_box.on_submit(D_submit)

        # making a text entry for alpha that allows for user input
        Avalue = "{alpha:.4e}".format(alpha=alpha)
        axIntAlpha = plt.axes([0.45, 0.075, 0.08, 0.03])
        Alpha_box = TextBox(axIntAlpha, '', initial=Avalue)
        Alpha_box.on_submit(A_submit)

        # making a text entry for n1 that allows for user input
        n1value = "{}".format(n1)
        # plt.figtext(0.0085, 0.178, "First:")
        # axIntn1 = plt.axes([0.075, 0.170, 0.08, 0.03])
        plt.figtext(0.02, 0.178, "First:")
        axIntn1 = plt.axes([0.05, 0.170, 0.05, 0.03])
        n1_box = TextBox(axIntn1, '', initial=n1)
        n1_box.on_submit(n1_submit)

        # making a text entry for n2 that allows for user input
        n2value = "{}".format(n2)
        # plt.figtext(0.17, 0.178, "Last:")
        # axIntn2 = plt.axes([0.235, 0.170, 0.08, 0.03])
        plt.figtext(0.12, 0.178, "Last:")
        axIntn2 = plt.axes([0.15, 0.170, 0.05, 0.03])
        n2_box = TextBox(axIntn2, '', initial=n2)
        n2_box.on_submit(n2_submit)

        # making a text entry for removing points that allows for user input
        nRemovevalue = "{}".format(nRemove)
        plt.figtext(0.21, 0.178, "Remove:")
        axnRemove = plt.axes([0.27, 0.170, 0.08, 0.03])
        nRemove_box = TextBox(axnRemove, '', initial=nRemove)
        nRemove_box.on_submit(nRemove_submit)

        # create a checkbox for extrapolation
        axExtrap = plt.axes([0.37, 0.170, 0.015, 0.03], frameon=True)
        axExtrap.margins(0.0)

        extrapolate_check = CheckButtons(axExtrap, ["Extrapolate"], [args.extrapolate])
        check = extrapolate_check
        try:
            #the axes object for the checkbutton is crazy large, and actually
            #blocks the sliders underneath even when frameon=False
            #so we have to manually set the size and location of each of the
            #elements of the checkbox after setting the axes margins to zero above
            #including the rectangle checkbox, the lines for the X, and the label
            size =  1.0 #size relative to axes axExtrap
            for rect in extrapolate_check.rectangles:
                rect.set_x(0.)
                rect.set_y(0.)
                rect.set_width(size)
                rect.set_height(size)
            first = True
            for l in check.lines:
                for ll in l:
                    llx = ll.get_xdata()
                    lly = ll.get_ydata()
                    #print(llx)
                    #print(lly)
                    ll.set_xdata([0.0,size])
                    if first:
                        #there's two lines making
                        #up the checkbox, so need
                        #to set the y values separately
                        #one going from bottom left to
                        #upper right, the other opposite
                        ll.set_ydata([size,0.0])
                        first = False
                    else:
                        ll.set_ydata([0.0, size])
        except:
            #newer versions of matplotlib don't have the check.rectangles attribute. so for now they get a small box.
            pass

        check.labels[0].set_position((1.5,.5))

        #here is the slider updating
        sdmax.on_changed(update)
        salpha.on_changed(update)
        extrapolate_check.on_clicked(update)
        #snes.on_changed(update)

        axreset = plt.axes([0.05, 0.02, 0.1, 0.04])
        reset_button = Button(axreset, 'Reset Sliders', color=axcolor, hovercolor='0.975')

        def reset_values(event):
            sdmax.reset()
            salpha.reset()
        reset_button.on_clicked(reset_values)

        axprint = plt.axes([0.2, 0.02, 0.1, 0.04])
        print_button = Button(axprint, 'Print Values', color=axcolor, hovercolor='0.975')

        print_button.on_clicked(print_values)

        axsave = plt.axes([0.35, 0.02, 0.1, 0.04])
        save_button = Button(axsave, 'Save File', color=axcolor, hovercolor='0.975')

        save_button.on_clicked(save_file)

        # plt.show()
### DENSS.bin.denss.fit_data.py copy  & modify  END ###

    return reset_button, print_button, save_button



"""
    Theory.Anisotropy.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from .JsPedersen1997 import F1, P6
from molass_legacy.Saxs.SaxsSamples import EllipsoidVoxels
from molass_legacy.Saxs.ReciprocalData import ReciprocalData
import molass_legacy.KekLib.DebugPlot as plt

class Anisotropy:
    def __init__(self):
        pass

    def fit_curve(self, q, y, Rg):
        def get_curve(p):
            a, b, c = p
            ellipsoid = EllipsoidVoxels(radii=(a, b, c))
            data = ellipsoid.get_data(density=0.1)
            rdata = ReciprocalData(data.shape)
            F = rdata.get_reciprocal(data)
            ey = rdata.get_scattering_curve(q, F)
            ey /= ey[0]
            return ey

        def plot_curve(ey):
            plt.push()
            fig, ax = plt.subplots()
            ax.set_yscale('log')
            ax.plot(q, y)
            ax.plot(q, ey)
            plt.show()
            plt.pop()

        def f(p):
            ey = get_curve(p)
            if False:
                plot_curve(ey)
            return np.sum((ey - y)**2)

        # R = np.sqrt(5/3)*Rg
        R = 0.2
        cons = (
            {'type': 'ineq', 'fun': lambda x:  x[1] - x[0]},
            {'type': 'ineq', 'fun': lambda x:  x[2] - x[1]},
        )
        bnds = ((0, None), (0, None), (0, None))

        dmin = None
        pmin = None
        for p in np.random.uniform(0, 1, (200,3)):
            sp = sorted(p)
            d = f(sp)
            if dmin is None or d < dmin:
                dmin = d
                pmin = sp

        print("pmin=", pmin)
        ey = get_curve(pmin)
        plot_curve(ey)

        if False:
            res = minimize(f, pmin, bounds=bnds, constraints=cons)
            self.params = res.x
            print('params=', self.params)

    def get_params(self):
        return self.params

    def get_curve(self):
        pass

def ellipsoid_scale(Rg, a, b, c):
    r = np.sqrt((a**2 + b**2 + c**2)/5)
    scale = Rg/r
    return scale

def demo():

    Rg = 35
    qv = np.linspace(0.005, 0.4, 100)

    R = np.sqrt(5/3)*Rg
    b1 = np.pi/R
    b2 = b1*1.5
    # a, b, c = np.random.uniform(0, 1, 3)
    a, b, c = 40, 35, 30
    print((a, b, c))
    scale = ellipsoid_scale(Rg, a, b, c)
    a_ = a*scale
    b_ = b*scale
    c_ = c*scale
    y_ = P6(qv, a_, b_, c_)

    ai = Anisotropy()
    ai.fit_curve(qv, y_, Rg)

    fig, ax = plt.subplots()

    ax.set_title("Anisotroy Estimation", fontsize=20)
    ax.set_yscale('log')
    ax.plot(qv, F1(qv, R)**2, label='Sphere')
    ax.plot(qv, y_, label='Ellipsoid(%.3g, %.3g, %.3g)' % (a_, b_, c_))

    ymin, ymax = ax.get_ylim()
    ax.set_ylim(ymin, ymax)
    for b, pc in zip([b1, b2], [50, 1]):
        ax.plot([b, b], [ymin, ymax], ':', color='red', label='rank2 ratio %d%%' % pc)

    ax.legend()
    fig.tight_layout()
    plt.show()



"""
    _MOLASS.WorkUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
from .SerialSettings import get_setting

def get_temp_folder(make_folder=True):
    temp_folder = get_setting('temp_folder')

    if temp_folder is None:
        analysis_folder = get_setting('analysis_folder')
        assert analysis_folder is not None, 'analysis_folder is not set'
        temp_folder = os.path.join(analysis_folder, '.temp')

    if make_folder:
        if not os.path.exists(temp_folder):
            os.makedirs(temp_folder)
    else:
        assert os.path.exists(temp_folder), 'temp_folder does not exist'

    return temp_folder



import numpy as np
# import matplotlib.pyplot as plt
import seaborn
seaborn.set()

# see https://github.com/sartorius-research/STLC for installation
from stlc import grm

import os
import sys
lib_dir = os.path.dirname(os.path.dirname(__file__))
sys.path.append(lib_dir)
kek_lib = os.path.join(lib_dir, "KekLib")
sys.path.append(kek_lib)
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.ElutionModels import emg
from SecTheory.BasicModels import fit_single_pore, robust_single_pore_pdf

def main():
    cm = 1e-2
    minute = 60
    zl = 1.7 * cm  #[cm]
    rp = 0.004 * cm  #[cm]
    ec = 0.4
    ep = 0.333
    kf = 0.01 * cm  #[cm]
    Dax = 0.002 * cm**2 / minute  #[cm^2 min^-1]
    Dp = 3.003 * 1e-6 * cm**2 / minute  #[cm^2 min^-1]
    Ds = 0.  #[cm^2 min^-1]
    u = 0.6 * cm / minute  #[cm min^-1]
    ka = 2.5  #[s^-1]
    kd = 1  #[s^-1]
    tinj = 20 * minute  #[min]
    cinj = 1  #[mol /m^-3]
    tmax = 100 * minute  #[min]
    qm = 1.
    nr = 10
    nz = 20
    dt = 1

    def step(t: float, tinj: float) -> float:
        return float(t <= tinj)

    parameters0 = grm.ModelParameters(c0=cinj,
                                          Dax=Dax,
                                          Dp=Dp,
                                          Ds=Ds,
                                          ka=ka,
                                          kd=kd,
                                          kf=kf,
                                          qm=qm,
                                          ip=lambda t: step(t, tinj))

    model = grm.GeneralRateModel(u=u,
                                         ep=ep,
                                         ec=ec,
                                         zl=zl,
                                         rp=rp,
                                         nz=nz,
                                         nr=nr,
                                         component_parameters=[parameters0])
    sol = grm.solve(model, tmax, dt)

    x = sol.t / minute
    y = sol.c[0, -1, :]

    print(len(sol.t), sol.t[0], sol.t[-1])

    m1 = np.sum(x*y)/np.sum(y)
    print("m1=", m1)
    m2 = np.sum((x - m1)**2 * y)/np.sum(y)
    print("m2=", m2)
    m3 = np.sum((x - m1)**3 * y)/np.sum(y)
    print("m3=", m3)

    tau = np.power(m3/2, 1/3)
    sigma = np.sqrt(m2 - tau**2)
    mu = m1 - tau

    y_ = emg(x, 1, mu, sigma, tau)
    y_ *= np.max(y)/np.max(y_)

    h, t0, np_, tp_ = fit_single_pore(x, y, moments=[m1, m2, m3])
    y_mp = h * robust_single_pore_pdf(x - t0, np_, tp_)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Comparison among GRM, EMG and Monopore")
        ax.plot(x, y, label='GRM - General Rate Model')
        ax.plot(x, y_, label='EMG')
        ax.plot(x, y_mp, label='Monopore')

        ax.legend()
        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    main()



"""

    CormapVersion.py

"""
import platform

def get_version_string(cpuid=False):
    if cpuid:
        from MachineTypes import get_cpuid
        cpuid = ' cpuid:' + str(get_cpuid())
    else:
        cpuid = ''

    return 'Cormap Maker 0.0.5 (2021-12-22 python %s %s%s)' % (
                platform.python_version(), platform.architecture()[0], cpuid )
